// Copyright (C) 2020 Intel Corporation

#pragma once

#include <string>
#include <map>
#include <memory>
#include <vector>
#include <numeric>
#include <curl/curl.h>

class RemoteInferenceEngine;

namespace InferenceEngine {

#define _CONFIG_KEY(name) KEY_##name
#define DECLARE_CONFIG_KEY(name) static constexpr auto _CONFIG_KEY(name) = #name
#define DECLARE_CONFIG_VALUE(name) static constexpr auto name = #name
#define CONFIG_KEY(name) InferenceEngine::PluginConfigParams::_CONFIG_KEY(name)
#define CONFIG_VALUE(name) InferenceEngine::PluginConfigParams::name

#define DECLARE_METRIC_KEY(name, ...)            \
	static constexpr auto METRIC_##name = #name
namespace Metrics {
DECLARE_METRIC_KEY(FULL_DEVICE_NAME, std::string);
};
#define METRIC_KEY(name) InferenceEngine::Metrics::METRIC_##name


namespace PluginConfigParams {
DECLARE_CONFIG_VALUE(YES);
DECLARE_CONFIG_VALUE(NO);
};

namespace GNAConfigParams {

#define GNA_CONFIG_KEY(name) InferenceEngine::GNAConfigParams::_CONFIG_KEY(GNA_##name)
#define GNA_CONFIG_VALUE(name) InferenceEngine::GNAConfigParams::GNA_##name
#define DECLARE_GNA_CONFIG_KEY(name) DECLARE_CONFIG_KEY(GNA_##name)
#define DECLARE_GNA_CONFIG_VALUE(name) DECLARE_CONFIG_VALUE(GNA_##name)

DECLARE_GNA_CONFIG_KEY(SCALE_FACTOR);
DECLARE_GNA_CONFIG_KEY(PRECISION);
DECLARE_GNA_CONFIG_KEY(DEVICE_MODE);
DECLARE_GNA_CONFIG_KEY(LIB_N_THREADS);
DECLARE_GNA_CONFIG_KEY(COMPACT_MODE);

DECLARE_GNA_CONFIG_VALUE(AUTO);
DECLARE_GNA_CONFIG_VALUE(HW);
DECLARE_GNA_CONFIG_VALUE(SW);
};

enum StatusCode : int {
	OK = 0, GENERAL_ERROR = -1, NOT_IMPLEMENTED = -2,
	NETWORK_NOT_LOADED = -3, PARAMETER_MISMATCH = -4,
	NOT_FOUND = -5, OUT_OF_BOUNDS = -6,
	UNEXPECTED = -7, REQUEST_BUSY = -8,
	RESULT_NOT_READY = -9, NOT_ALLOCATED = -10,
	INFER_NOT_STARTED = -11, NETWORK_NOT_READ = -12
};

enum Precision : uint8_t {
	UNSPECIFIED = 255, MIXED = 0, FP32 = 10, FP16 = 11,
	BF16 = 12, Q78 = 20, I16 = 30, U8 = 40,
	I8 = 50, U16 = 60, I32 = 70, I64 = 72,
	U64 = 73, BIN = 71, BOOL = 41, CUSTOM = 80
};

enum Layout : uint8_t {
	ANY = 0, NCHW = 1, NHWC = 2, NCDHW = 3, NDHWC = 4,
	OIHW = 64, GOIHW = 65, OIDHW = 66, GOIDHW = 67, SCALAR = 95,
	C = 96,
	CHW = 128,
	HW = 192, NC = 193, CN = 194,
	BLOCKED = 200,
};


using SizeVector = std::vector<size_t>;

class TensorDesc {
public:
	Precision precision;
	SizeVector dims;
	Layout layout;
	TensorDesc() {};
	TensorDesc(Precision precision, const SizeVector& dims, Layout layout)
		: precision(precision), dims(dims), layout(layout) {};

	const SizeVector&  getDims() const { return dims; };
	SizeVector&  getDims() { return dims; };
	Layout getLayout() const { return layout; };
	Precision getPrecision() const { return precision; };
};

template <class T>
class LockedMemory {
public:
	LockedMemory(T *p) { buffer = p; };
	T *buffer;
	template<class S> S as() { return (S)(buffer); };
	operator T*() { return buffer; };
	operator const T*() const { return buffer; };
};

class Blob {
public:
	virtual ~Blob() {};
	using Ptr = std::shared_ptr<Blob>;
	using CPtr = std::shared_ptr<const Blob>;
	virtual size_t size() const = 0;
	virtual size_t byteSize() const = 0;
	virtual LockedMemory<void> buffer() noexcept = 0;
	virtual LockedMemory<const void> cbuffer() const noexcept = 0;

	static size_t product(const SizeVector& dims) noexcept {
		if (dims.empty()) return 0;
		return std::accumulate(std::begin(dims), std::end(dims),
				       (size_t)1, std::multiplies<size_t>());
	}
};

class MemoryBlob: public Blob {
protected:
	void initBuffer(size_t);
public:
	char *buf;
	size_t bufferSize;
	MemoryBlob(size_t size);
	MemoryBlob(): buf(NULL), bufferSize(0) {} ;
	~MemoryBlob();
	using Ptr = std::shared_ptr<MemoryBlob>;
	using CPtr = std::shared_ptr<const MemoryBlob>;
	LockedMemory<void> wmap();
	LockedMemory<const void> rmap() const;

	virtual LockedMemory<void> buffer() noexcept override;
	virtual LockedMemory<const void> cbuffer() const noexcept override;
	virtual size_t byteSize() const;
	virtual size_t size() const;
};

template <typename T>
class TBlob : public MemoryBlob {
public:
	TensorDesc tdesc;
	using Ptr = std::shared_ptr<TBlob<T>>;

	TBlob(const TensorDesc& tdesc, T* ptr, size_t data_size = 0):
		tdesc(tdesc) {
		if (data_size == 0)
			data_size = size();
		if (data_size != 0 && ptr == nullptr)
			throw "Using Blob on external nullptr memory";

		initBuffer(data_size);
		memcpy(buf, ptr, data_size);
	}

	virtual size_t size() const {
		if (tdesc.getLayout() == Layout::SCALAR) return 1;
		return product(tdesc.getDims());
	}

	virtual size_t byteSize() const {
		return size() * sizeof(T);
	}
};

template <typename Type>
inline typename InferenceEngine::TBlob<Type>::Ptr
make_shared_blob(const TensorDesc& tdesc, Type* ptr, size_t size = 0) {
	//if (!tdesc.getPrecision().hasStorageType<Type>())
		//throw("Cannot make shared blob! "
		      //"The blob type cannot be used to "
		      //"store objects of current precision");
	return std::make_shared<InferenceEngine::TBlob<Type>>(tdesc, ptr, size);
}

class InputInfo {
public:
	size_t size;
	TensorDesc tdesc;
	void setPrecision(Precision precision);
	const TensorDesc& getTensorDesc() const { return tdesc; };

	using Ptr = std::shared_ptr<InputInfo>;
	using CPtr = std::shared_ptr<const InputInfo>;
};
using ConstInputsDataMap = std::map<std::string, InputInfo::CPtr>;
using InputsDataMap = std::map<std::string, InputInfo::Ptr>;


class Data {
public:
	std::string name;
	TensorDesc tdesc;
	size_t size;
	void setPrecision(Precision precision) {};
	const TensorDesc& getTensorDesc() const { return tdesc; };
	const SizeVector& getDims() const { return tdesc.getDims(); };
	const std::string& getName() const { return name; };
};
using DataPtr = std::shared_ptr<Data>;
using CDataPtr = std::shared_ptr<const Data>;
using OutputsDataMap = std::map<std::string, DataPtr>;
using ConstOutputsDataMap = std::map<std::string, CDataPtr>;

class IInferRequest {
public:
	enum WaitMode : int64_t {
		RESULT_READY = -1,
		STATUS_ONLY = 0,
	};
};

class MemoryState {
public:
	void Reset(){};
};

class InferRequest {
public:
	std::shared_ptr<RemoteInferenceEngine> rie;
	Blob::Ptr GetBlob(const std::string &name);
	void Infer();
	StatusCode Wait(int64_t millis_timeout);
	std::vector<MemoryState> QueryState();
};

class ExecutableNetwork {
public:
	ConstInputsDataMap inputsInfo;
	ConstOutputsDataMap outputsInfo;

	std::shared_ptr<RemoteInferenceEngine> rie;
	InferRequest CreateInferRequest();
	ConstInputsDataMap GetInputsInfo();
	ConstOutputsDataMap GetOutputsInfo();
	std::vector<MemoryState> QueryState();
};

class CNNNetwork {
public:
	std::shared_ptr<RemoteInferenceEngine> rie;
	InputsDataMap inputsInfo;
	OutputsDataMap outputsInfo;
	CNNNetwork(const std::string &modelPath);
	CNNNetwork();
	void setBatchSize(int);
	InputsDataMap getInputsInfo();
	OutputsDataMap getOutputsInfo();
};

struct Version {
	struct {
		int major;
		int minor;
	} apiVersion;
	const char *buildNumber;
	const char *description;
};

class Parameter {
public:
	// now support std::string only
	std::string value;
	template <typename T> T as() { return value; };
};

class Core {
public:
	Core();
	~Core();
	CNNNetwork ReadNetwork(const std::string &model,
			       const Blob::CPtr &weights) const;
	CNNNetwork ReadNetwork(const std::string &modelPath,
			       const std::string &binPath="");
	ExecutableNetwork LoadNetwork(const CNNNetwork &network,
				      const std::string &device);
	ExecutableNetwork LoadNetwork(const CNNNetwork &network,
				      const std::string &device,
				      const std::map<std::string, std::string> &config);
	ExecutableNetwork ImportNetwork(const std::string &modelPath,
					const std::string &device,
					const std::map<std::string, std::string> &config = {});
	ExecutableNetwork ImportNetwork(std::istream &networkModel,
					const std::string &device,
					const std::map<std::string, std::string> &config = {});
	std::map<std::string, Version> GetVersions(const std::string &device) const;

	std::vector<std::string> GetAvailableDevices() const;
	Parameter GetMetric(const std::string &deviceName, const std::string &name) const;
};

template <typename T>
std::shared_ptr<T> as(const Blob::Ptr& blob) noexcept {
	return std::dynamic_pointer_cast<T>(blob);
}

const Version* GetInferenceEngineVersion();

};
